@@ -10,6 +10,7 @@ use crate::error::*;
 use crate::sess::{Session, SessionIo};
 use crate::src::{SourceFile, SourceGroup};
 use crate::target::{TargetSet, TargetSpec};
+use common_path::common_path_all;
 
 use std::collections::HashSet;
 
@@ -40,6 +41,7 @@ pub fn new<'a, 'b>() -> App<'a, 'b> {
                     "genus",
                     "vivado",
                     "vivado-sim",
+                    "precision",
                 ]),
         )
         .arg(
@@ -124,13 +126,48 @@ pub fn new<'a, 'b>() -> App<'a, 'b> {
                     "common",
                 ])
         )
+        .arg(
+            Arg::with_name("package")
+                .short("p")
+                .long("package")
+                .help("Specify package to show sources for")
+                .takes_value(true)
+                .multiple(true)
+                .number_of_values(1),
+        )
+        .arg(
+            Arg::with_name("no_deps")
+                .short("n")
+                .long("no-deps")
+                .help("Exclude all dependencies, i.e. only top level or specified package(s)"),
+        )
+        .arg(
+            Arg::with_name("exclude")
+                .short("e")
+                .long("exclude")
+                .help("Specify package to exclude from sources")
+                .takes_value(true)
+                .multiple(true)
+                .number_of_values(1),
+        )
+}
+
+fn get_package_strings<I>(packages: I) -> HashSet<String>
+where
+    I: IntoIterator,
+    I::Item: AsRef<str>,
+{
+    packages
+        .into_iter()
+        .map(|t| t.as_ref().to_string())
+        .collect()
 }
 
 /// Execute the `script` subcommand.
 pub fn run(sess: &Session, matches: &ArgMatches) -> Result<()> {
     let mut core = Core::new().unwrap();
     let io = SessionIo::new(&sess, core.handle());
-    let srcs = core.run(io.sources())?;
+    let mut srcs = core.run(io.sources())?;
 
     // Format-specific target specifiers.
     let vivado_targets = &["vivado", "fpga", "xilinx"];
@@ -148,6 +185,7 @@ pub fn run(sess: &Session, matches: &ArgMatches) -> Result<()> {
         "genus" => vec!["genus", "synthesis"],
         "vivado" => concat(vivado_targets, &["synthesis"]),
         "vivado-sim" => concat(vivado_targets, &["simulation"]),
+        "precision" => vec!["precision", "fpga", "synthesis"],
         _ => unreachable!(),
     };
 
@@ -162,17 +200,51 @@ pub fn run(sess: &Session, matches: &ArgMatches) -> Result<()> {
         .values_of("target")
         .map(|t| TargetSet::new(t.chain(format_targets.clone())))
         .unwrap_or_else(|| TargetSet::new(format_targets));
-    let srcs = srcs
+    srcs = srcs
         .filter_targets(&targets)
         .unwrap_or_else(|| SourceGroup {
             package: Default::default(),
             independent: true,
             target: TargetSpec::Wildcard,
             include_dirs: Default::default(),
+            export_incdirs: Default::default(),
             defines: Default::default(),
             files: Default::default(),
+            dependencies: Default::default(),
         });
 
+    // Filter the sources by specified packages.
+    let packages = &srcs.get_package_list(
+        sess,
+        &matches
+            .values_of("package")
+            .map(|p| get_package_strings(p))
+            .unwrap_or_else(|| HashSet::new()),
+        &matches
+            .values_of("exclude")
+            .map(|p| get_package_strings(p))
+            .unwrap_or_else(|| HashSet::new()),
+        matches.is_present("no_deps"),
+    );
+
+    if matches.is_present("package")
+        || matches.is_present("exclude")
+        || matches.is_present("no_deps")
+    {
+        srcs = srcs
+            .filter_packages(&packages)
+            .unwrap_or_else(|| SourceGroup {
+                package: Default::default(),
+                independent: true,
+                target: TargetSpec::Wildcard,
+                include_dirs: Default::default(),
+                export_incdirs: Default::default(),
+                defines: Default::default(),
+                files: Default::default(),
+                dependencies: Default::default(),
+            });
+    }
+
     // Flatten the sources.
     let srcs = srcs.flatten();
 
@@ -215,6 +287,7 @@ pub fn run(sess: &Session, matches: &ArgMatches) -> Result<()> {
         "genus" => emit_genus_tcl(sess, matches, targets, srcs),
         "vivado" => emit_vivado_tcl(sess, matches, targets, srcs),
         "vivado-sim" => emit_vivado_tcl(sess, matches, targets, srcs),
+        "precision" => emit_precision_tcl(sess, matches, targets, srcs, abort_on_error),
         _ => unreachable!(),
     }
 }
@@ -362,7 +435,7 @@ fn emit_vsim_tcl(
                             }
                             lines.push(s);
                         }
-                        for i in &src.include_dirs {
+                        for i in src.clone().get_incdirs() {
                             lines
                                 .push(quote(&format!("+incdir+{}", relativize_path(i, sess.root))));
                         }
@@ -447,7 +520,7 @@ fn emit_vcs_sh(
                             }
                             lines.push(s);
                         }
-                        for i in &src.include_dirs {
+                        for i in src.clone().get_incdirs() {
                             lines
                                 .push(quote(&format!("+incdir+{}", relativize_path(i, sess.root))));
                         }
@@ -520,7 +593,7 @@ fn emit_verilator_sh(
                             }
                             lines.push(s);
                         }
-                        for i in &src.include_dirs {
+                        for i in src.clone().get_incdirs() {
                             if i.starts_with(sess.root) {
                                 lines.push(format!(
                                     "+incdir+{}/{}",
@@ -560,17 +633,10 @@ fn emit_verilator_sh(
 /// Emit a flat file list
 fn emit_flist(sess: &Session, matches: &ArgMatches, srcs: Vec<SourceGroup>) -> Result<()> {
     let mut lines = vec![];
-    let mut inc_dirs = HashSet::new();
+    let mut inc_dirs = Vec::new();
     let mut files = vec![];
-    // Gobble double includes with a HashSet.
     for src in srcs {
-        inc_dirs = src
-            .include_dirs
-            .into_iter()
-            .fold(HashSet::new(), |mut acc, inc_dir| {
-                acc.insert(inc_dir);
-                acc
-            });
+        inc_dirs = src.clone().get_incdirs();
         files.append(&mut src.files.clone());
     }
 
@@ -625,7 +691,7 @@ fn emit_synopsys_tcl(
         // Adjust the search path.
         println!("");
         println!("set search_path $search_path_initial");
-        for i in &src.include_dirs {
+        for i in src.clone().get_incdirs() {
             println!("lappend search_path {}", relativize_path(i));
         }
 
@@ -738,7 +804,7 @@ fn emit_genus_tcl(
         // Adjust the search path.
         println!("");
         println!("set search_path $search_path_initial");
-        for i in &src.include_dirs {
+        for i in src.clone().get_incdirs() {
             println!("lappend search_path {}", relativize_path(i));
         }
 
@@ -859,7 +925,7 @@ fn emit_vivado_tcl(
         vec!["", " -simset"]
     };
     for src in srcs {
-        for i in &src.include_dirs {
+        for i in src.clone().get_incdirs() {
             include_dirs.push(relativize_path(i, sess.root));
         }
         separate_files_in_group(
@@ -983,7 +1049,7 @@ fn emit_riviera_tcl(
                                 }
                                 lines.push(s);
                             }
-                            for i in &src.include_dirs {
+                            for i in src.clone().get_incdirs() {
                                 lines.push(quote(&format!(
                                     "+incdir+{}",
                                     relativize_path(i, sess.root)
@@ -1015,18 +1081,12 @@ fn emit_riviera_tcl(
     } else {
         let mut lines = vec![];
         let mut file_lines = vec![];
-        let mut inc_dirs = HashSet::new();
+        let mut inc_dirs = Vec::new();
         let mut files = vec![];
         let mut defines: Vec<(String, Option<String>)> = vec![];
         let mut t: bool = false;
         for src in srcs {
-            inc_dirs = src
-                .include_dirs
-                .into_iter()
-                .fold(HashSet::new(), |mut acc, inc_dir| {
-                    acc.insert(inc_dir);
-                    acc
-                });
+            inc_dirs = src.clone().get_incdirs();
             files.append(&mut src.files.clone());
             defines.extend(
                 src.defines
@@ -1088,3 +1148,117 @@ fn emit_riviera_tcl(
     }
     Ok(())
 }
+
+/// Emit a script to add sources to Mentor Precision
+fn emit_precision_tcl(
+    sess: &Session,
+    matches: &ArgMatches,
+    targets: TargetSet,
+    srcs: Vec<SourceGroup>,
+    abort_on_error: bool,
+) -> Result<()> {
+    // Find the common_path between session and all source files
+    let mut file_paths = vec![sess.root];
+    for src in &srcs {
+        for file in &src.files {
+            let p = match file {
+                SourceFile::File(p) => p,
+                _ => continue,
+            };
+            file_paths.push(p)
+        }
+    }
+    let root = common_path_all(file_paths).unwrap();
+
+    // Print the script header
+    println!("{}", format!("# {}", HEADER_AUTOGEN));
+    println!("# Precision does not take relative paths into account when specifying include dirs.");
+    println!("# Define the common ROOT anyway if needed for patching file paths. ");
+    println!("set ROOT {}", root.to_str().unwrap());
+    println!("set_input_dir $ROOT");
+    println!("setup_design -search_path $ROOT");
+
+    // Find all the include dirs as precision only allows to set these globally
+    let mut defines: Vec<(String, Option<String>)> = vec![];
+    for src in &srcs {
+        defines.extend(
+            src.defines
+                .iter()
+                .map(|(k, &v)| (k.to_string(), v.map(String::from))),
+        );
+    }
+    defines.extend(
+        targets
+            .iter()
+            .map(|t| (format!("TARGET_{}", t.to_uppercase()), None)),
+    );
+    add_defines_from_matches(&mut defines, matches);
+    defines.sort();
+    if !defines.is_empty() {
+        let mut lines = vec![];
+
+        lines.push("setup_design -defines { \\".to_owned());
+        for (k, v) in defines {
+            let mut s = format!("    +define+{}", k);
+            if let Some(v) = v {
+                s.push('=');
+                s.push_str(&v);
+            }
+            s.push_str(" \\");
+            lines.push(s);
+        }
+        lines.push("}".to_owned());
+        println!("\n# Set globally all defines for the (S)Verilog sources.");
+        println!("{}", lines.join("\n"));
+    }
+
+    // Add the source files depending on group
+    for src in srcs {
+        separate_files_in_group(
+            src,
+            |f| match f {
+                SourceFile::File(p) => match p.extension().and_then(std::ffi::OsStr::to_str) {
+                    Some("sv") | Some("v") | Some("vp") => Some(SourceType::Verilog),
+                    Some("vhd") | Some("vhdl") => Some(SourceType::Vhdl),
+                    _ => None,
+                },
+                _ => None,
+            },
+            |src, ty, files| {
+                let mut lines = vec![];
+                match ty {
+                    SourceType::Verilog => {
+                        lines.push(tcl_catch_prefix("add_input_file", abort_on_error).to_owned());
+                        lines.push("-format SystemVerilog2012".to_owned());
+                        if !src.clone().get_incdirs().is_empty() {
+                            lines.push("-search_path {".to_owned());
+                            for i in src.clone().get_incdirs() {
+                                lines.push(format!("    {}", i.to_str().unwrap()));
+                            }
+                            lines.push("}".to_owned());
+                        }
+                    }
+                    SourceType::Vhdl => {
+                        lines.push(tcl_catch_prefix("add_input_file", abort_on_error).to_owned());
+                        lines.push("-format vhdl_2008".to_owned());
+                    }
+                }
+                lines.push("{".to_owned());
+                for file in files {
+                    let p = match file {
+                        SourceFile::File(p) => p,
+                        _ => continue,
+                    };
+                    lines.push(format!("    {}", p.to_str().unwrap()));
+                }
+                lines.push("} \\".to_owned());
+                println!("");
+                println!("{}", lines.join(" \\\n    "));
+                if abort_on_error {
+                    println!("{}", tcl_catch_postfix());
+                }
+            },
+        );
+    }
+    Ok(())
+}
