@@ -9,6 +9,7 @@ use std;
 use std::collections::{HashMap, HashSet};
 use std::fmt;
 use std::fs::canonicalize;
+use std::iter::FromIterator;
 use std::mem::swap;
 use std::path::{Path, PathBuf};
 use std::sync::atomic::AtomicUsize;
@@ -70,6 +71,8 @@ pub struct Session<'ctx> {
     pub cache: SessionCache<'ctx>,
     /// A throttle for futures performing git network operations.
     git_throttle: FutureThrottle,
+    /// A toggle to disable remote fetches & clones
+    pub local_only: bool,
 }
 
 impl<'sess, 'ctx: 'sess> Session<'ctx> {
@@ -79,13 +82,21 @@ impl<'sess, 'ctx: 'sess> Session<'ctx> {
         manifest: &'ctx Manifest,
         config: &'ctx Config,
         arenas: &'ctx SessionArenas,
+        local_only: bool,
+        force_fetch: bool,
     ) -> Session<'ctx> {
         Session {
             root: root,
             manifest: manifest,
             config: config,
             arenas: arenas,
-            manifest_mtime: try_modification_time(root.join("Bender.yml")),
+            manifest_mtime: {
+                if force_fetch {
+                    Some(SystemTime::now())
+                } else {
+                    try_modification_time(root.join("Bender.yml"))
+                }
+            },
             stats: Default::default(),
             deps: Mutex::new(DependencyTable::new()),
             paths: Mutex::new(HashSet::new()),
@@ -97,6 +108,7 @@ impl<'sess, 'ctx: 'sess> Session<'ctx> {
             plugins: Mutex::new(None),
             cache: Default::default(),
             git_throttle: FutureThrottle::new(8),
+            local_only: local_only,
         }
     }
 
@@ -138,7 +150,7 @@ impl<'sess, 'ctx: 'sess> Session<'ctx> {
     ///
     /// This internalizes the dependency sources, i.e. assigns `DependencyRef`
     /// objects to them, and generates a nametable.
-    pub fn load_locked(&self, locked: &config::Locked) {
+    pub fn load_locked(&self, locked: &config::Locked) -> Result<()> {
         let mut deps = self.deps.lock().unwrap();
         let mut names = HashMap::new();
         let mut graph_names = HashMap::new();
@@ -178,7 +190,17 @@ impl<'sess, 'ctx: 'sess> Session<'ctx> {
             let mut ranks: HashMap<DependencyRef, usize> =
                 graph.keys().map(|&id| (id, 0)).collect();
             let mut pending = HashSet::new();
+            for name in self.manifest.dependencies.keys() {
+                if !(names.contains_key(name)) {
+                    return Err(Error::new(format!(
+                        "`Bender.yml` contains dependency `{}` but `Bender.lock` does not.\n\
+                        \tYou may need to run `bender update`.",
+                        name
+                    )));
+                }
+            }
             pending.extend(self.manifest.dependencies.keys().map(|name| names[name]));
+            let mut cyclic = false;
             while !pending.is_empty() {
                 let mut current_pending = HashSet::new();
                 swap(&mut pending, &mut current_pending);
@@ -190,6 +212,22 @@ impl<'sess, 'ctx: 'sess> Session<'ctx> {
                             pending.insert(dep_id);
                         }
                     }
+                    // Limit rank to two times graph length, which is sufficient except if there is
+                    // a cyclic dependency
+                    if ranks[&id] > 2 * graph.len() {
+                        cyclic = true;
+                    }
+                }
+                if cyclic == true {
+                    let mut pend_str = vec![];
+                    for element in pending.iter() {
+                        pend_str.push(self.dependency_name(*element));
+                    }
+                    return Err(Error::new(format!(
+                        "a cyclical dependency was discovered, likely relates to one of {:?}.\n\
+                        \tPlease ensure no dependency loops.",
+                        pend_str
+                    )));
                 }
             }
             debugln!("sess: topological ranks {:#?}", ranks);
@@ -218,6 +256,7 @@ impl<'sess, 'ctx: 'sess> Session<'ctx> {
         *self.names.lock().unwrap() = names;
         *self.graph.lock().unwrap() = Arc::new(graph);
         *self.pkgs.lock().unwrap() = Arc::new(pkgs);
+        Ok(())
     }
 
     /// Obtain information on a dependency.
@@ -324,12 +363,11 @@ impl<'sess, 'ctx: 'sess> Session<'ctx> {
         &self,
         sources: &'ctx config::Sources,
         package: Option<&'ctx str>,
+        dependencies: Vec<String>,
+        dependency_export_includes: HashMap<String, Vec<&'ctx Path>>,
     ) -> SourceGroup<'ctx> {
-        let include_dirs = sources
-            .include_dirs
-            .iter()
-            .map(|d| self.intern_path(d))
-            .collect();
+        let include_dirs: HashSet<&Path> =
+            HashSet::from_iter(sources.include_dirs.iter().map(|d| self.intern_path(d)));
         let defines = sources
             .defines
             .iter()
@@ -345,18 +383,25 @@ impl<'sess, 'ctx: 'sess> Session<'ctx> {
             .iter()
             .map(|file| match *file {
                 config::SourceFile::File(ref path) => (path as &Path).into(),
-                config::SourceFile::Group(ref group) => {
-                    self.load_sources(group.as_ref(), None).into()
-                }
+                config::SourceFile::Group(ref group) => self
+                    .load_sources(
+                        group.as_ref(),
+                        None,
+                        dependencies.clone(),
+                        dependency_export_includes.clone(),
+                    )
+                    .into(),
             })
             .collect();
         SourceGroup {
             package: package,
             independent: false,
             target: sources.target.clone(),
-            include_dirs: include_dirs,
+            include_dirs: include_dirs.into_iter().collect(),
+            export_incdirs: dependency_export_includes.clone(),
             defines: defines,
             files: files,
+            dependencies: dependencies,
         }
     }
 }
@@ -453,6 +498,12 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
 
         // Either initialize the repository or update it if needed.
         if !db_dir.join("config").exists() {
+            if self.sess.local_only {
+                return Box::new(future::err(Error::new(format!(
+                    "Bender --local argument set, unable to initialize git dependency. \n\
+                    \tPlease update without --local, or provide a path to the missing dependency."
+                ))));
+            }
             // Initialize.
             self.sess.stats.num_database_init.increment();
             Box::new(
@@ -478,7 +529,7 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
         } else {
             // Update if the manifest has been modified since the last fetch.
             let db_mtime = try_modification_time(db_dir.join("FETCH_HEAD"));
-            if self.sess.manifest_mtime < db_mtime && !force_fetch {
+            if (self.sess.manifest_mtime < db_mtime && !force_fetch) || self.sess.local_only {
                 debugln!("sess: skipping fetch of {:?}", db_dir);
                 return Box::new(future::ok(git));
             }
@@ -613,17 +664,8 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
         }
     }
 
-    /// Ensure that a dependency is checked out and obtain its path.
-    pub fn checkout(
-        &'io self,
-        dep_id: DependencyRef,
-    ) -> Box<dyn Future<Item = &'ctx Path, Error = Error> + 'io> {
-        // Check if the checkout is already in the cache.
-        if let Some(&cached) = self.sess.cache.checkout.lock().unwrap().get(&dep_id) {
-            return Box::new(future::ok(cached));
-        }
-
-        self.sess.stats.num_calls_checkout.increment();
+    /// Get the path of a dependency
+    fn get_package_path(&'io self, dep_id: DependencyRef) -> PathBuf {
         let dep = self.sess.dependency(dep_id);
 
         // Determine the name of the checkout as the given name and the first
@@ -644,8 +686,7 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
                         Ok(p) => p,
                         Err(_) => path,
                     };
-                    let path = self.sess.intern_path(path);
-                    return Box::new(future::ok(path));
+                    return path.to_path_buf();
                 }
             }
             hasher.update(format!("{:?}", self.sess.root).as_bytes());
@@ -657,16 +698,43 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
         // explicit checkout directory, use that and do not append any hash to
         // the dependency name.
         let checkout_dir = match self.sess.manifest.workspace.checkout_dir {
-            Some(ref cd) => self.sess.intern_path(cd.join(&dep.name)),
-            None => self.sess.intern_path(
-                self.sess
-                    .config
-                    .database
-                    .join("git")
-                    .join("checkouts")
-                    .join(checkout_name),
-            ),
+            Some(ref cd) => cd.join(&dep.name),
+            None => self
+                .sess
+                .config
+                .database
+                .join("git")
+                .join("checkouts")
+                .join(checkout_name),
         };
+        checkout_dir.to_path_buf()
+    }
+
+    /// Ensure that a dependency is checked out and obtain its path.
+    pub fn checkout(
+        &'io self,
+        dep_id: DependencyRef,
+    ) -> Box<dyn Future<Item = &'ctx Path, Error = Error> + 'io> {
+        // Check if the checkout is already in the cache.
+        if let Some(&cached) = self.sess.cache.checkout.lock().unwrap().get(&dep_id) {
+            return Box::new(future::ok(cached));
+        }
+
+        self.sess.stats.num_calls_checkout.increment();
+        let dep = self.sess.dependency(dep_id);
+
+        match dep.source {
+            DependencySource::Registry => unimplemented!(),
+            DependencySource::Git(..) => {}
+            DependencySource::Path(..) => {
+                let path = self
+                    .sess
+                    .intern_path(self.get_package_path(dep_id).as_path());
+                return Box::new(future::ok(path));
+            }
+        }
+
+        let checkout_dir = self.sess.intern_path(self.get_package_path(dep_id));
 
         match dep.source {
             DependencySource::Path(..) => unreachable!(),
@@ -824,10 +892,19 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
         use self::DependencyVersion as DepVer;
         match (&dep.source, version) {
             (&DepSrc::Path(ref path), DepVer::Path) => {
+                if !path.starts_with("/") {
+                    warnln!("There may be issues in the path for {:?}.", dep.name);
+                }
                 let manifest_path = path.join("Bender.yml");
                 if manifest_path.exists() {
                     match read_manifest(&manifest_path) {
-                        Ok(m) => Box::new(future::ok(Some(self.sess.intern_manifest(m)))),
+                        Ok(m) => {
+                            if dep.name != m.package.name {
+                                warnln!("Dependency name and package name do not match for {:?} / {:?}, this can cause unwanted behavior",
+                                    dep.name, m.package.name); // TODO: This should be an error
+                            }
+                            Box::new(future::ok(Some(self.sess.intern_manifest(m))))
+                        }
                         Err(e) => Box::new(future::err(e)),
                     }
                 } else {
@@ -866,7 +943,7 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
                                             cause,
                                         )
                                     })?;
-                                let full = partial.validate().map_err(|cause| {
+                                let mut full = partial.validate().map_err(|cause| {
                                     Error::chain(
                                         format!(
                                             "Error in manifest of dependency `{}` at revision \
@@ -876,6 +953,20 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
                                         cause,
                                     )
                                 })?;
+                                // Add base path to path dependencies within git repositories
+                                for dep in full.dependencies.iter_mut() {
+                                    match dep {
+                                        (_, config::Dependency::Path(ref path)) => {
+                                            if !path.starts_with("/") {
+                                                if !self.get_package_path(dep_id).exists() {
+                                                    warnln!("Please note that dependencies for {:?} may not be available unless {:?} is properly checked out.\n         (to checkout run `bender sources` and then `bender update` again).", dep.0, full.package.name);
+                                                }
+                                                *dep.1 = config::Dependency::Path(self.get_package_path(dep_id).join(path).clone());
+                                            }
+                                        },
+                                        (_, _) => {},
+                                    }
+                                }
                                 Ok(Some(self.sess.intern_manifest(full)))
                             }
                             None => Ok(None),
@@ -887,6 +978,16 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
                                 .lock()
                                 .unwrap()
                                 .insert(cache_key, manifest);
+                            if dep.name != match manifest {
+                                Some(x) => &x.package.name,
+                                None => "dead"
+                            } {
+                                warnln!("Dependency name and package name do not match for {:?} / {:?}, this can cause unwanted behavior",
+                                    dep.name, match manifest {
+                                        Some(x) => &x.package.name,
+                                        None => "dead"
+                                    }); // TODO (micprog): This should be an error
+                            }
                             Ok(manifest)
                         }),
                 )
@@ -975,26 +1076,39 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
         Box::new(
             manifests
                 .and_then(move |ranks| {
-                    // Collect a set of exported include directories that will be applied to
-                    // all packages.
-                    // TODO: Make this more fine-grained, such that only include dirs from
-                    // directly depended-on packages are considered.
-                    let export_include_dirs: Vec<_> = ranks
-                        .iter()
-                        .chain(once(&vec![Some(self.sess.manifest)]))
-                        .flat_map(|manifests| {
+                    use std::iter::once;
+
+                    // Create HashMap of the export_include_dirs for each package
+                    let mut all_export_include_dirs: HashMap<String, Vec<&Path>> = HashMap::new();
+                    let tmp_export_include_dirs: Vec<HashMap<_, _>> = ranks
+                        .clone()
+                        .into_iter()
+                        .chain(once(vec![Some(self.sess.manifest)]))
+                        .map(|manifests| {
                             manifests
-                                .iter()
-                                .filter_map(|&m| m)
-                                .flat_map(|m| m.export_include_dirs.iter())
+                                .clone()
+                                .into_iter()
+                                .filter_map(|m| m)
+                                .map(|m| {
+                                    (
+                                        m.package.name.clone(),
+                                        m.export_include_dirs
+                                            .iter()
+                                            .map(PathBuf::as_path)
+                                            .collect(),
+                                    )
+                                })
+                                .collect()
                         })
                         .collect();
+                    for element in tmp_export_include_dirs {
+                        all_export_include_dirs.extend(element);
+                    }
                     debugln!(
-                        "explicitly exported include dirs: {:#?}",
-                        export_include_dirs
+                        "export_include_dirs for each package: {:?}",
+                        all_export_include_dirs
                     );
 
-                    use std::iter::once;
                     let files = ranks
                         .into_iter()
                         .chain(once(vec![Some(self.sess.manifest)]))
@@ -1004,8 +1118,36 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
                                 .filter_map(|m| m)
                                 .filter_map(|m| {
                                     m.sources.as_ref().map(|s| {
+                                        // Collect include dirs from export_include_dirs of package and direct dependencies
+                                        let mut export_include_dirs: HashMap<String, Vec<&Path>> =
+                                            HashMap::new();
+                                        export_include_dirs.insert(
+                                            m.package.name.clone(),
+                                            m.export_include_dirs
+                                                .iter()
+                                                .map(PathBuf::as_path)
+                                                .collect(),
+                                        );
+                                        if !m.dependencies.is_empty() {
+                                            for i in m.dependencies.keys() {
+                                                if !all_export_include_dirs.contains_key(i) {
+                                                    warnln!("Name issue with {:?}, `export_include_dirs` not handled\n\tCould relate to name mismatch, see `bender update`", i);
+                                                    export_include_dirs.insert(i.clone(), Vec::new());
+                                                } else {
+                                                    export_include_dirs.insert(
+                                                        i.clone(),
+                                                        all_export_include_dirs[i].clone(),
+                                                    );
+                                                }
+                                            }
+                                        }
                                         self.sess
-                                            .load_sources(s, Some(m.package.name.as_str()))
+                                            .load_sources(
+                                                s,
+                                                Some(m.package.name.as_str()),
+                                                m.dependencies.keys().cloned().collect(),
+                                                export_include_dirs,
+                                            )
                                             .into()
                                     })
                                 })
@@ -1017,8 +1159,10 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
                                 independent: true,
                                 target: TargetSpec::Wildcard,
                                 include_dirs: Vec::new(),
+                                export_incdirs: HashMap::new(),
                                 defines: HashMap::new(),
                                 files: files,
+                                dependencies: Vec::new(),
                             }
                             .into()
                         })
@@ -1030,12 +1174,11 @@ impl<'io, 'sess: 'io, 'ctx: 'sess> SessionIo<'sess, 'ctx> {
                         package: None,
                         independent: false,
                         target: TargetSpec::Wildcard,
-                        include_dirs: export_include_dirs
-                            .iter()
-                            .map(|p| self.sess.intern_path(p.clone()))
-                            .collect(),
+                        include_dirs: Vec::new(),
+                        export_incdirs: HashMap::new(),
                         defines: HashMap::new(),
                         files: files,
+                        dependencies: Vec::new(),
                     }
                     .simplify())
                 })
